/*
 * Copyright (c) 1995 - 2001 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <xfs/xfs_locl.h>
#include <xfs/xfs_common.h>
#include <xfs/xfs_fs.h>
#include <xfs/xfs_deb.h>

RCSID("$Id: xfs_node-osf.c,v 1.14.2.4 2001/06/27 15:11:26 mattiasa Exp $");

extern struct vnodeops *xfs_vnodeop_p;
extern struct vfs_ubcops xfs_ubcops;

/*
 * Create a new xfs_node and make a VN_HOLD()!
 *
 * Also prevents creation of duplicates. This happens
 * whenever there are more than one name to a file,
 * "." and ".." are common cases.
 */

int
new_xfs_node(struct xfs *xfsp,
	     struct xfs_msg_node *node,
	     struct xfs_node **xpp,
	     struct proc *p)
{
    int do_vget = 0;
    int do_ins = 0;
    struct xfs_node *result;

    XFSDEB(XDEBNODE, ("new_xfs_node %d.%d.%d.%d\n",
		      node->handle.a,
		      node->handle.b,
		      node->handle.c,
		      node->handle.d));

    /* Does not allow duplicates */
    result = xfs_node_find(xfsp, &node->handle);

    if (result == 0) {
	int error;
	struct vnode *v;

	error = getnewvnode(VT_AFS, &xfs_vnodeops, &v);
	    XFSDEB(XDEBVNOPS,
		   ("new_xfs_node: v = %lx v->v_usecount = %d\n", 
		    (unsigned long)v, v?v->v_usecount:0));

	if (error) {
	    XFSDEB(XDEBVNOPS,
		   ("XFS PANIC! new_xfs_node: could not allocate node"));
	    return error;
	}
	do_ins++; /* do insmntque below */
	result = VNODE_TO_XNODE(v);
	bzero(result, sizeof(*result));
	result->vn = v;

	result->anonrights = node->anonrights;

	result->handle = node->handle;
	result->flags = 0;
	result->tokens = 0;

	xfsp->nnodes++;
    } else {
	/* Node is already cached */
	do_vget = 1;
    }

    /* Init other fields */
    xfs_attr2vattr(&node->attr, &result->attr, 1);
    result->vn->v_type = result->attr.va_type;
    XFS_TOKEN_SET(result, XFS_ATTR_R, XFS_ATTR_MASK);
    bcopy(node->id, result->id, sizeof(result->id));
    bcopy(node->rights, result->rights, sizeof(result->rights));

    if(do_ins)
	insmntque(result->vn, XFS_TO_VFS(xfsp), &xfs_ubcops); /* not done by 
								 getnewvnode */
    /*
     * We need to postpone this until here because (on FreeBSD) vget
     * tries to install a pager on the vnode and for that it wants to
     * retrieve the size with getattr.
     */

    if(do_vget) {
	vget(XNODE_TO_VNODE(result)); /* XXX unlock? */
    }
    *xpp = result;
    XFSDEB(XDEBNODE, ("return: new_xfs_node\n"));
    return 0;
}

void
free_xfs_node(struct xfs_node *node)
{
    struct xfs *xfsp = XFS_FROM_XNODE(node);

    XFSDEB(XDEBNODE, ("free_xfs_node starting: node = %lx\n",
		      (unsigned long)node));

    /* XXX Really need to put back dirty data first. */

    if (DATA_FROM_XNODE(node)) {
	vrele(DATA_FROM_XNODE(node));
	DATA_FROM_XNODE(node) = NULL;
    }
    xfsp->nnodes--;

    XFSDEB(XDEBNODE, ("free_xfs_node done\n"));
}

int
free_all_xfs_nodes(struct xfs *xfsp, int flags, int unmountp)
{
    int error = 0;
    struct mount *mp = XFS_TO_VFS(xfsp);

    if (mp == NULL) {
	XFSDEB(XDEBNODE, ("free_all_xfs_nodes already freed\n"));
	return 0;
    }

    XFSDEB(XDEBNODE, ("free_all_xfs_nodes starting\n"));

    xfs_dnlc_purge_mp(mp);

    XFSDEB(XDEBNODE, ("free_all_xfs_nodes now removing root\n"));

    if (xfsp->root) {
	struct vnode *vp = XNODE_TO_VNODE(xfsp->root);
	XFSDEB(XDEBNODE, ("root->v_usecount = %d\n", vp->v_usecount));
	VN_LOCK(vp);
	vgone(vp, VX_NOSLEEP, NULL);
	VN_UNLOCK(vp);
	vrele(vp);
	xfsp->root = 0;
    }

    XFSDEB(XDEBNODE, ("free_all_xfs_nodes root removed\n"));

    XFSDEB(XDEBNODE, ("free_all_xfs_nodes now killing all remaining nodes\n"));

    error = vflush(mp, NULL, flags);

    if (error) {
	XFSDEB(XDEBNODE, ("xfree_all_xfs_nodes: vflush() error == %d\n",
			  error));
	return error;
    }

    XFSDEB(XDEBNODE, ("free_all_xfs_nodes done\n"));
    return error;
}

struct xfs_node *
xfs_node_find(struct xfs *xfsp, xfs_handle *handlep)
{
    struct vnode *t;
    struct xfs_node *xn = NULL;

    XFSDEB(XDEBNODE, ("xfs_node_find: xfsp = %lx handlep = %lx\n", 
		      (unsigned long)xfsp,
		      (unsigned long)handlep));

    for(t = XFS_TO_VFS(xfsp)->m_mounth; 
	t != NULL; 
	t = t->v_mountf) {
	xn = VNODE_TO_XNODE(t);
	if (xn && xfs_handle_eq(&xn->handle, handlep))
	    break;
    }

    if (t != NULL)
	return xn;
    else
	return NULL;
}

void
vattr2xfs_attr(const struct vattr *va, struct xfs_attr *xa)
{
    bzero(xa, sizeof(*xa));
    if (va->va_mode != (u_short)VNOVAL)
	XA_SET_MODE(xa, va->va_mode);
    if (va->va_nlink != VNOVAL)
	XA_SET_NLINK(xa, va->va_nlink);
    if (va->va_size != VNOVAL)
	XA_SET_SIZE(xa, va->va_size);
    if (va->va_uid != (uid_t)VNOVAL)
	XA_SET_UID(xa, va->va_uid);
    if (va->va_gid != (gid_t)VNOVAL)
	XA_SET_GID(xa, va->va_gid);
    if (va->va_atime.tv_sec != VNOVAL)
	XA_SET_ATIME(xa, va->va_atime.tv_sec);
    if (va->va_mtime.tv_sec != VNOVAL)
	XA_SET_MTIME(xa, va->va_mtime.tv_sec);
    if (va->va_ctime.tv_sec != VNOVAL)
	XA_SET_CTIME(xa, va->va_ctime.tv_sec);
    if (va->va_fileid != VNOVAL)
	XA_SET_FILEID(xa, va->va_fileid);
    switch (va->va_type) {
    case VNON:
	xa->xa_type = XFS_FILE_NON;
	break;
    case VREG:
	xa->xa_type = XFS_FILE_REG;
	break;
    case VDIR:
	xa->xa_type = XFS_FILE_DIR;
	break;
    case VBLK:
	xa->xa_type = XFS_FILE_BLK;
	break;
    case VCHR:
	xa->xa_type = XFS_FILE_CHR;
	break;
    case VLNK:
	xa->xa_type = XFS_FILE_LNK;
	break;
    case VSOCK:
	xa->xa_type = XFS_FILE_SOCK;
	break;
    case VFIFO:
	xa->xa_type = XFS_FILE_FIFO;
	break;
    case VBAD:
	xa->xa_type = XFS_FILE_BAD;
	break;
    default:
	panic("xfs_attr2attr: bad value");
    }
}

#define SET_TIMEVAL(X, S, N) do { (X)->tv_sec = (S); (X)->tv_usec = (N) / 1000; } while(0)

void
xfs_attr2vattr(const struct xfs_attr *xa, struct vattr *va, int clear_node)
{
    if (clear_node)
	VATTR_NULL(va);
    if (XA_VALID_MODE(xa))
	va->va_mode = xa->xa_mode;
    if (XA_VALID_NLINK(xa))
	va->va_nlink = xa->xa_nlink;
    if (XA_VALID_SIZE(xa))
	va->va_size = xa->xa_size;
    if (XA_VALID_UID(xa))
	va->va_uid = xa->xa_uid;
    if (XA_VALID_GID(xa))
	va->va_gid = xa->xa_gid;
    if (XA_VALID_ATIME(xa)) {
	SET_TIMEVAL(&va->va_atime, xa->xa_atime, 0);
    }
    if (XA_VALID_MTIME(xa)) {
	SET_TIMEVAL(&va->va_mtime, xa->xa_mtime, 0);
    }
    if (XA_VALID_CTIME(xa)) {
	SET_TIMEVAL(&va->va_ctime, xa->xa_ctime, 0);
    }
    if (XA_VALID_FILEID(xa)) {
	va->va_fileid = xa->xa_fileid;
    }
    if (XA_VALID_TYPE(xa)) {
	switch (xa->xa_type) {
	case XFS_FILE_NON:
	    va->va_type = VNON;
	    break;
	case XFS_FILE_REG:
	    va->va_type = VREG;
	    break;
	case XFS_FILE_DIR:
	    va->va_type = VDIR;
	    break;
	case XFS_FILE_BLK:
	    va->va_type = VBLK;
	    break;
	case XFS_FILE_CHR:
	    va->va_type = VCHR;
	    break;
	case XFS_FILE_LNK:
	    va->va_type = VLNK;
	    break;
	case XFS_FILE_SOCK:
	    va->va_type = VSOCK;
	    break;
	case XFS_FILE_FIFO:
	    va->va_type = VFIFO;
	    break;
	case XFS_FILE_BAD:
	    va->va_type = VBAD;
	    break;
	default:
	    panic("xfs_attr2vattr: bad value");
	}
    }
    va->va_flags = 0;
    va->va_blocksize = 8192;
    va->va_bytes = va->va_size;
}

/*
 * A single entry DNLC for systems for handling long names that don't
 * get put into the system DNLC.
 */

struct long_entry {
    struct vnode *dvp, *vp;
    char name[MAXNAMLEN + 1];
    size_t len;
    u_long dvpid, vpid;
};

static struct long_entry tbl;

/*
 * Nuke the `tbl'
 */

static void
tbl_clear (void)
{
    tbl.dvp = tbl.vp = NULL;
    tbl.name[0] = '\0';
    tbl.len = 0;
    tbl.dvpid = tbl.vpid = 0;
}

/*
 * Set the entry in the `tbl'
 */

static void
tbl_enter (size_t len, const char *name, struct vnode *dvp, struct vnode *vp)
{
    tbl.len = len;
    bcopy(name, tbl.name, len);
    tbl.dvp = dvp;
    tbl.vp = vp;
    tbl.dvpid = dvp->v_id;
    tbl.vpid = vp->v_id;
}

/*
 * Lookup in tbl
 */

static int
tbl_lookup (xfs_componentname *cnp, struct vnode *dvp,
	    struct vnode **res)
{
    size_t len = cnp->cn_namelen;
    const char *name = cnp->cn_nameptr;

    if (tbl.dvp == dvp
	&& tbl.len == len
	&& strncmp(tbl.name, name, len) == 0
	&& tbl.dvpid == tbl.dvp->v_id
	&& tbl.vpid == tbl.vp->v_id) {
	*res = tbl.vp;
	CACHE_LOOKUP_REF(*res);
#if	UNIX_LOCKS
	cnp->ni_vpid = (*res)->v_id;
#endif
	return -1;
    } else
	return 0;
}

/*
 * Store a nameidata in the DNLC
 */

int
xfs_dnlc_enter(struct vnode *dvp,
	       xfs_componentname *ndp,
	       struct vnode *vp)
{
    XFSDEB(XDEBDNLC, ("xfs_dnlc_enter(%lx, %lx, %lx)\n",
		      (unsigned long)dvp,
		      (unsigned long)ndp,
		      (unsigned long)vp));
    XFSDEB(XDEBDNLC, ("xfs_dnlc_enter: v_id = %ld\n", dvp->v_id));

    XFSDEB(XDEBDNLC, ("xfs_dnlc_enter: calling cache_enter:"
		      "dvp = %lx, vp = %lx, ndp = (%s, %ld, %lu, %lx, %lx)\n",
		      (unsigned long)dvp,
		      (unsigned long)vp, ndp->ni_ptr, ndp->ni_namelen,
		      ndp->cn_hash, ndp->ni_ptr, ndp->ni_next));

    if (ndp->ni_namelen <= NCHNAMLEN)
    {
	/*
	 * This is to make sure there's no negative entry already in the dnlc
	 */
	short save_nameiop;
	struct vnode *dummy;
	int ret;

	ndp->ni_dvp = dvp;

	save_nameiop = ndp->ni_nameiop;
	ndp->ni_nameiop = CREATE;
	ndp->ni_makeentry = 1;

	if ((ret = cache_lookup(ndp)) != 0) {
	    printf ("XFS PANIC WARNING! xfs_dnlc_enter: "
		    "%s already in cache, ret = %d\n",
		    ndp->ni_ptr, ret);
	}

	if (ndp->ni_vp == NULL && vp != NULL)
	    panic ("xfs_dnlc_enter: bad result from cache_lookup");

	ndp->ni_nameiop = save_nameiop;

	XFSDEB(XDEBDNLC, ("xfs_dnlc_enter: entering cache_enter\n"));
	cache_enter(ndp);
	XFSDEB(XDEBDNLC, ("xfs_dnlc_enter: leaving cache_enter\n"));
    }

    if (vp != NULL)
	tbl_enter (ndp->ni_namelen, ndp->ni_ptr, dvp, vp);

    return 0;
}

static void
xfs_cnp_init (struct nameidata *ndp,
	      const char *name, struct vnode *vp, struct vnode *dvp,
	      struct proc *proc, struct ucred *cred,
	      int nameiop)
{
    int i;

    ndinit(ndp,NULL);

    ndp->ni_ptr = (char *)name;
    ndp->ni_namelen = strlen(name);
    ndp->ni_next = ndp->ni_ptr + ndp->ni_namelen;
    ndp->ni_vp = vp;
    ndp->ni_dvp = dvp;
    ndp->ni_hash = 0;

    if (ndp->ni_namelen <= NCHNAMLEN) {


#ifdef HAVE_FOUR_ARGUMENT_VFS_NAME_HASH
	vfs_name_hash(ndp->ni_ptr, ndp->ni_dent.d_name,
		      dvp, &ndp->ni_hash);
#elif HAVE_THREE_ARGUMENT_VFS_NAME_HASH
	vfs_name_hash(ndp->ni_ptr, ndp->ni_dent.d_name,
		      &ndp->ni_hash);
#else
	for (i = 0; i < ndp->ni_namelen; ++i)
	    ndp->ni_hash += name[i] * (i + 1);
#endif
    }

    ndp->ni_nameiop = nameiop;
    ndp->ni_cred = cred;
}

/*
 * Store a (dvp, name, vp) in the DNLC
 */

int
xfs_dnlc_enter_name(struct vnode *dvp,
		    const char *name,
		    struct vnode *vp)
{
    struct nameidata nd;

    XFSDEB(XDEBDNLC, ("xfs_dnlc_enter_name(%lx, \"%s\", %lx)\n",
		      (unsigned long)dvp, name,
		      (unsigned long)vp));

    xfs_cnp_init (&nd, name, vp, dvp, NULL, NULL, CREATE);
    return xfs_dnlc_enter (dvp, &nd, vp);
}

 /*
  * do the last (and locking protocol) portion of xnlc_lookup
  *
  * return:
  * -1 for succesful
  * 0  for failed
  */

static int
xfs_dnlc_lock(struct vnode *dvp,
	      xfs_componentname *cnp,
	      struct vnode **res)
{
    int error = 0, vpid;

#if	UNIX_LOCKS
    vpid = cnp->ni_vpid;
#else
    vpid = (*res)->v_id;
#endif

    if (dvp == *res) {		/* "." */
	VREF(dvp);
	abort_vget(dvp);
	if (vpid != (*res)->v_id) {
	    error = 1;
	    vrele(*res);
	}
    } else {
	VN_LOCK(*res);
	error = vget_cache_nowait(*res);
	if (!error) {
	    if (vpid != (*res)->v_id) {
		error = 1;
		VN_UNLOCK(*res);
		vrele(*res);
	    } else
		VN_UNLOCK(*res);
	} else
	    VN_UNLOCK(*res);
    }

    if(error) {
	*res = NULL;
	return 0;
    } else
	return -1;
}

/*
 * Lookup (dvp, cnp) in the DNLC and return the result in `res'.
 * Return -1 if succesful, 0 if not and ENOENT if we're sure it doesn't exist.
 */

int
xfs_dnlc_lookup(struct vnode *dvp,
		xfs_componentname *cnp,
		struct vnode **res)
{
    int error;

    XFSDEB(XDEBDNLC, ("xfs_dnlc_lookup(%lx, \"%s\")\n",
		      (unsigned long)dvp, cnp->cn_nameptr));
    
    XFSDEB(XDEBDNLC, ("xfs_dnlc_lookup: v_id = %ld\n", dvp->v_id));
    
    XFSDEB(XDEBDNLC, ("xfs_dnlc_lookup: calling cache_lookup:"
		      "dvp = %lx, cnp = (%s, %ld, %lu)\n",
		      (unsigned long)dvp,
		      cnp->cn_nameptr, cnp->cn_namelen, cnp->cn_hash));

    cnp->ni_makeentry = 1;

    cnp->ni_dvp = dvp;
    error = cache_lookup(cnp);

#if 0
    *res = cnp->ni_vp;
#endif

    XFSDEB(XDEBDNLC, ("xfs_dnlc_lookup: cache_lookup returned. "
		      "error = %d, *res = %lx res->v_usecount = %d\n", error,
		      (unsigned long)*res, *res?(*res)->v_usecount:0));

    if (error == 0)
	error = tbl_lookup (cnp, dvp, res);

    XFSDEB(XDEBDNLC, ("xfs_dnlc_lookup: tbl_lookup returned. "
		      "error = %d, *res = %lx res->v_usecount = %d\n", error,
		      (unsigned long)*res, *res?(*res)->v_usecount:0));


    if (error != -1) {
	*res = NULL;
	return error;
    }

    return xfs_dnlc_lock (dvp, cnp, res);
}

/*
 *
 */

int
xfs_dnlc_lookup_name (struct vnode *dvp,
		      const char *name,
		      struct vnode **res)
{
    struct nameidata nd;

    xfs_cnp_init (&nd, name, NULL, dvp, NULL, NULL, LOOKUP);
    return xfs_dnlc_lookup (dvp, &nd, res);
}


/*
 * Remove one entry from the DNLC
 */

void
xfs_dnlc_purge (struct vnode *vp)
{
    XFSDEB(XDEBDNLC, ("xfs_dnlc_purge\n"));

    if (tbl.dvp == vp || tbl.vp == vp)
	tbl_clear ();

    cache_purge(vp);
}

/*
 * Remove all entries belong to `mp' from the DNLC
 */

void
xfs_dnlc_purge_mp (struct mount *mp)
{
    XFSDEB(XDEBDNLC, ("xfs_dnlc_purge_mp()\n"));

    tbl_clear ();
#ifdef HAVE_KERNEL_CACHE_PURGEVFS
    cache_purgevfs(mp);
#else
    /* XXX */
#endif /* HAVE_KERNEL_CACHE_PURGEVFS */
}

/*
 * Returns 1 if pag has any rights set in the node
 */

int
xfs_has_pag(const struct xfs_node *xn, xfs_pag_t pag)
{
    int i;

    for (i = 0; i < MAXRIGHTS; i++)
	if (xn->id[i] == pag)
	    return 1;

    return 0;
}
